from datetime import datetime

from sqlalchemy.ext.asyncio import AsyncEngine

from aiogram import BaseMiddleware
from aiogram.enums import ChatType, ContentType, ChatMemberStatus
from aiogram.types import TelegramObject, User, Message, CallbackQuery, InlineQuery, ChatMemberUpdated

from .models import BaseModel, ChatModel, UserModel, MessageModel, CallbackQueryModel, InlineQueryModel
from ..sqlalchemy.session import AsyncSessionManagerFactory
from ..utils.logging import loggers

class AnalyticsMiddleware(BaseMiddleware):
	"""
	Middleware to collect analytics from user interactions.

	:param path: Path to SQL database file
	:param chats: Whether to track chats
	:param messages: Whether to track messages
	:param callback_queries: Whether to track callback queries
	:param inline_queries: Whether to track inline queries
	"""

	def __init__(self, engine: str | AsyncEngine, chats: bool, messages: bool, callback_queries: bool, inline_queries: bool):
		self.tables_created = False

		self.engine = engine
		self.session_manager_factory = AsyncSessionManagerFactory(self.engine)

		self.chats = chats
		self.messages = messages
		self.inline_queries = inline_queries
		self.callback_queries = callback_queries

	async def create_all(self):
		if self.tables_created:
			loggers.analytics.error("Tables are already created")

			return False
		
		async with self.engine.begin() as connection:
			await connection.run_sync(BaseModel.metadata.create_all)

		self.tables_created = True

		return True

	async def __call__(self, handler, event: TelegramObject, data: dict):
		user: User = data.get("event_from_user")
		
		if user:
			async with self.session_manager_factory() as session:
				is_new = await session.get(UserModel, user.id) is None

				if is_new:
					loggers.analytics.info(f"New user id={user.id}")

				model = UserModel(
					id=user.id,
					username=user.username,
					first_name=user.first_name,
					last_name=user.last_name,
					full_name=user.full_name,
					is_premium=bool(user.is_premium)
				)

				if is_new:
					model.joined_at = datetime.now()

				await session.merge(model)

			if isinstance(event, ChatMemberUpdated):
				if event.chat.type == ChatType.PRIVATE and event.bot.id in (event.old_chat_member.user.id, event.new_chat_member.user.id):
					if event.new_chat_member.status == ChatMemberStatus.MEMBER:
						blocked = False
					elif event.new_chat_member.status == ChatMemberStatus.KICKED:
						blocked = True
					else:
						blocked = None

					if blocked is not None:
						if blocked:
							loggers.analytics.info(f"Blocked by user id={user.id}")
						else:
							loggers.analytics.info(f"Unblocked by user id={user.id}")

						async with self.session_manager_factory() as session:
							(await session.get(UserModel, user.id)).blocked = blocked
				
				if self.chats:
					loggers.analytics.info("chats")
					async with self.session_manager_factory() as session:
						is_new = await session.get(ChatModel, event.chat.id) is None

						model = ChatModel(
							id=event.chat.id,
							type=event.chat.type,
							title=event.chat.title,
							full_name=event.chat.full_name,
							username=event.chat.username,
							bio=event.chat.bio
						)

						if is_new:
							model.joined_at = event.date

						await session.merge(model)
		
			if isinstance(event, Message) and self.messages:
				loggers.analytics.info(f"New message from user id={user.id}")

				async with self.session_manager_factory() as session:
					if event.content_type == ContentType.PHOTO:
						file_id = event.photo[-1].file_id
					elif event.content_type in [
						ContentType.DOCUMENT,
						ContentType.VIDEO,
						ContentType.AUDIO,
						ContentType.VOICE,
						ContentType.VIDEO_NOTE,
						ContentType.ANIMATION,
						ContentType.STICKER
					]:
						file_id = getattr(event, event.content_type).file_id
					else:
						file_id = None

					model = MessageModel(
						id=event.message_id,
						chat_id=event.chat.id,
						user_id=user.id,
						text=event.text or event.caption,
						md_text=event.md_text,
						html_text=event.html_text,
						entities=event.model_dump()["entities"],
						content_type=event.content_type,
						file_id=file_id,
						media_group_id=event.media_group_id,
						reply_to_message_id=event.reply_to_message.message_id if event.reply_to_message else None,
						date=event.date
					)

					session.add(model)

			if isinstance(event, CallbackQuery) and self.callback_queries:
				loggers.analytics.info(f"New callback query from user id={user.id}")

				async with self.session_manager_factory() as session:
					model = CallbackQueryModel(
						id=event.id,
						user_id=user.id,
						message_id=event.message.message_id,
						data=event.data,
						date=datetime.now()
					)

					session.add(model)

			if isinstance(event, InlineQuery) and self.inline_queries:
				loggers.analytics.info(f"New inline query from user id={user.id}")

				async with self.session_manager_factory() as session:
					model = InlineQueryModel(
						id=event.id,
						user=await session.get(UserModel, user.id),
						query=event.query,
						chat_type=event.chat_type,
						date=datetime.now()
					)

					session.add(model)

		return await handler(event, data)
	
__all__ = ["AnalyticsMiddleware"]


import logging
import sys

LOG_FORMAT = "\x1b[30;1m%(asctime)s\x1b[0m \x1b[34;1m%(levelname)s\x1b[0m \x1b[35m%(name)s\x1b[0m %(message)s"
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

RED = "\033[91m"
GREEN = "\033[92m"
YELLOW = "\033[93m"
BLUE = "\033[94m"
RESET = "\033[0m"

LEVEL_COLORS = {
	logging.DEBUG: GREEN,
	logging.INFO: BLUE,
	logging.WARNING: YELLOW,
	logging.ERROR: RED,
	logging.CRITICAL: RED,
}

class Formatter(logging.Formatter):
	def __init__(self, fmt, datefmt=None, style="%"):
		super().__init__(fmt, datefmt, style)

	def format(self, record: logging.LogRecord):
		record.levelname = LEVEL_COLORS.get(record.levelno, "") + record.levelname.ljust(8) + RESET
			
		return super().format(record)

class MuteLoggersFilter(logging.Filter):
	def __init__(self, *names: str):
		self.names = set(names)

		super().__init__()

	def filter(self, record):
		return record.name not in self.names

def setup_logging(level: int=logging.INFO, levels: dict[str, int | bool]=dict()):
	logging.captureWarnings(True)

	handler = logging.StreamHandler(sys.stdout)
	handler.setFormatter(Formatter(fmt=LOG_FORMAT, datefmt=DATE_FORMAT))
	
	root = logging.getLogger()
	root.setLevel(level)

	muted_loggers = []

	levels |= {
		"asyncio": False,
		"aiosqlite": False,
		"watchfiles.main": False
	}

	for name, level in levels.items():
		if level == False: # noqa E712
			muted_loggers.append(name)
		else:
			logging.getLogger(name).setLevel(level)

	handler.addFilter(MuteLoggersFilter(*muted_loggers))

	root.handlers.clear()
	root.addHandler(handler)

	logging.getLogger("tgtilities")

get_logger = logging.getLogger


import asyncio
import logging

from tgtilities.analytics import setup_analytics
from tgtilities.routers import include_routers
from tgtilities.sqlalchemy.session import AsyncSessionManagerFactory
from tgtilities.utils.logging import setup_logging

from aiogram import Bot, Dispatcher

setup_logging(logging.DEBUG, levels={
	"aiogram.event": logging.WARNING
})

session_manager_factory = AsyncSessionManagerFactory("users.db")

async def main():
	bot = Bot(token="7659457511:AAGbU2nydXarlXfacgOFg1iGA7XNK7IpFno")

	dispatcher = Dispatcher()

	await setup_analytics(dispatcher)
	include_routers(dispatcher, "handlers", hot_reload=True)
	
	await dispatcher.start_polling(bot, session_factory=session_manager_factory)

if __name__ == "__main__":
	asyncio.run(main())


import asyncio
import sys
from importlib.util import spec_from_file_location, module_from_spec
from pathlib import Path
from watchfiles import awatch, Change

from aiogram import Router

from ..utils.logging import loggers
	
def extract_router(file_path: str | Path) -> Router:
	"""
	Extract `router` from file at `file_path`.

	:param file_path: Path to Python file

	:rtype: Router
	:return: Module's `router` attribute

	:raises ModuleNotFoundError: If `spec_from_file_location` fails
	:raises TypeError: If `router` isn't an instance of `Router`
	"""
	
	file_path = Path(file_path)

	spec = spec_from_file_location(file_path.stem, file_path)

	if spec is None or spec.loader is None:
		raise ModuleNotFoundError(f"Failed to create spec from file location {str(file_path)!r}")

	module = module_from_spec(spec)
	module.__name__ = spec.name
	module.__file__ = str(file_path)

	sys.modules[spec.name] = module
	
	spec.loader.exec_module(module)

	router = getattr(module, "router", None)

	if not isinstance(router, Router):
		raise TypeError(f"Expected Router, got {type(router).__name__}")

	return router

def unload_router(router: Router) -> Router:
	"""
	Remove `router` from it's `parent_router`'s sub-routers.

	:param router: Router to unload

	:rtype: Router
	:return: The unloaded router
	"""

	if router.parent_router is not None:
		router.parent_router.sub_routers = [
			sub_router
			for sub_router in router.parent_router.sub_routers
			if sub_router.name != router.name
		]

	return router

def load_router(file_path: str | Path, parent_router: Router) -> Router:
	"""
	Extract router from `file_path` and attach it to `parent_router`.

	:param file_path: File path to extract router from
	:param parent_router: Router to parent to

	:rtype: Router
	:return: The loaded router
	"""
	
	router = extract_router(file_path)
	parent_router.include_router(router)

	return router

def reload_router(router: Router, file_path: str | Path) -> Router:
	"""
	Reload router by first unloading it, then loading a new one from `file_path`.

	:param router: Router to reload
	:param file_path: File path to extract router from

	:rtype: Router
	:return: The reloaded router
	"""

	unload_router(router)
	return load_router(file_path, router.parent_router)

def handle_router_naming(router: Router, routers: dict[Path, Router]):
	"""
	Renames `router` if one of `routers` has the same name.

	:param router: Router to handle naming of
	"""

	if getattr(router, "name", None) is None:
		return
	
	if router.name in [loaded_router.name for loaded_router in routers.values()]:
		old_name = router.name
		router.name = f"{old_name}_{hex(id(router))}"

		loggers.routers.warning(f"Renamed {old_name!r} to {router.name!r} due to name conflict")

async def awatch_routers(directory: str | Path, parent_router: Router, routers: dict[Path, Router]):
	"""
	Asynchronously watch files in `directory` and reload routers on change.

	:param directory: Path to routers directory
	:param parent_router: Router to attach routers from newly added files to
	:param routers: Dictionary with file paths mapped to already loaded routers
	"""

	async for changes in awatch(directory):
		for change_type, changed_path in changes:
			changed_path = Path(changed_path).resolve()

			router = routers.get(changed_path)

			if change_type == Change.modified:
				routers.pop(changed_path)
				
				try:
					router = reload_router(router, changed_path)
				except: # noqa E722
					if getattr(router, "name", None) is None:
						loggers.routers.error(f"Exception while reloading router from file {str(changed_path)!r}", exc_info=True)
					else:
						loggers.routers.error(f"Exception while reloading router {router.name!r}", exc_info=True)
				else:
					loggers.routers.info(f"Reloaded router {router.name!r}")
				finally:
					handle_router_naming(router, routers)

					routers[changed_path] = router

			elif change_type == Change.added:
				try:
					router = load_router(changed_path, parent_router)
				except: # noqa E722
					if getattr(router, "name", None) is None:
						loggers.routers.error(f"Exception while loading router from file {str(changed_path)!r}", exc_info=True)
					else:
						loggers.routers.error(f"Exception while loading router {router.name!r}", exc_info=True)
				else:
					loggers.routers.info(f"Loaded router {router.name!r}")
				finally:
					handle_router_naming(router, routers)

					routers[changed_path] = router
			elif change_type == Change.deleted:
				unload_router(router)
				routers.pop(changed_path)

				loggers.routers.info(f"Unloaded router {router.name!r}")

def include_routers(parent_router: Router, directory: str | Path, hot_reload: bool=False):
	"""
	Include routers from directory with optional hot-reloading

	:param router: Router (or Dispatcher) to attach routers to
	:param directory: Path to routers directory
	:param hot_reload: Whether to hot reload routers on changes
	"""

	directory = Path(directory).resolve()

	routers = dict()

	for file_path in directory.iterdir():
		if file_path.suffix == ".py":
			router = extract_router(file_path)

			if router.name is None:
				router.name = file_path.stem

			routers[file_path] = load_router(file_path, parent_router)

	if hot_reload:
		asyncio.create_task(awatch_routers(directory, parent_router, routers))

__all__ = ["extract_router", "unload_router", "load_router", "reload_router", "handle_router_naming", "awatch_routers", "include_routers"]


from aiogram.exceptions import TelegramBadRequest

class SupressTelegramBadRequest:
	def __init__(self, message: str | None=None):
		self.message = message

	def __enter__(self):
		return self
	
	def __exit__(self, exc_type: type[Exception], exc: Exception, tb):
		return (
			isinstance(exc, TelegramBadRequest)
			and self.message in exc.message
		)
	
class SupressNotModified(SupressTelegramBadRequest):
	def __init__(self):
		super().__init__("message is not modified")


class ReprMixin:
	"""
	Mixin with a simple `__repr__` implementation for SQLAlchemy models.
	"""

	def __repr__(self):
		attributes = ", ".join(
			f"{key}={value}"
			for key, value in vars(self).items()
			if not key.startswith("_") and not hasattr(value, "_sa_adapter") and not hasattr(value, "_sa_instance_state")
		)

		return f"<{self.__class__.__name__}({attributes})>"

__all__ = ["ReprMixin"]


import logging as _logging

_package_name = "tgtilities"

logging = _logging.getLogger(f"{_package_name}.logging")
routers = _logging.getLogger(f"{_package_name}.routers")
sqlalchemy = _logging.getLogger(f"{_package_name}.sqlalchemy")
analytics = _logging.getLogger(f"{_package_name}.analytics")

__all__ = ["logging", "routers", "analytics"]


from aiogram import Router
from aiogram.filters import CommandStart
from aiogram.types import Message

router = Router(name="start")

@router.message(CommandStart())
async def start(message: Message):
	await message.answer("Hello!")


from aiogram import Router
from aiogram.filters import Command
from aiogram.types import Message, ReplyKeyboardRemove

from tgtilities.keyboard import static_reply_markup, static_inline_markup 

router = Router(name="keyboards")

@static_reply_markup
def reply_markup():
	return [
		["Меню"],
		["Профиль", "Настройки"]
	]

@static_inline_markup
def inline_markup():
	return [
		[("Меню", "menu"), ("Товары", "products")],
		[("Настройки", "settings")]
	]

@router.message(Command("reply"))
async def reply(message: Message):
	await message.answer(
		"Выбери кнопку ниже 👇",
		reply_markup=reply_markup()
	)

@router.message(Command("inline"))
async def inline(message: Message):
	await message.answer(
		"Выбери инлайн-кнопку ниже 👇",
		reply_markup=inline_markup()
	)

@router.message(Command("remove"))
async def remove(message: Message):
	await message.answer(
		"Клавиатура убрана",
		reply_markup=ReplyKeyboardRemove()
	)


def truncate(string: str, left: int, right: int):
	"""
	Truncate `string` by keeping `left` characters from the start and `right` characters from the end.

	Example::

		truncate("Hajimemashite!", 4, 4)
		# Haji...ite!

	:param string: String to truncate
	:param left: Number of characters to keep from the start
	:param right: Number of characters to keep from the end

	:return: Truncated string
	:rtype: str
	"""

	return string[:left] + "..." + string[-right:]

def ltruncate(string: str, length: int, ellipsis: bool=True):
	"""
	Truncate `string` from the start.

	Example::

		truncate("Hajimemashite!", 4)
		# ...ite!

	:param string: String to truncate
	:param length: Number of characters to keep
	:param ellipsis: Whether to prepend an ellipsis

	:return: Truncated string
	:rtype: str
	"""

	result = string[-length:]
	
	if ellipsis:
		return "..." + result
	else:
		return result

def rtruncate(string: str, length: int, ellipsis: bool=True):
	"""
	Truncate `string` from the end.

	Example::

		truncate("Hajimemashite!", 4)
		# Haji...

	:param string: String to truncate
	:param length: Number of characters to keep
	:param ellipsis: Whether to append an ellipsis

	:return: Truncated string
	:rtype: str
	"""

	result = string[:length]
	
	if ellipsis:
		return result + "..."
	else:
		return result


from functools import wraps
from typing import Callable
from typing_extensions import ParamSpec

from aiogram.types import KeyboardButton, InlineKeyboardButton
from aiogram.utils.keyboard import ReplyKeyboardBuilder, InlineKeyboardBuilder

P = ParamSpec("P")

def reply(function: Callable[P, list[list[str]]]):
	@wraps(function)
	def wrapper(*args: P.args, **kwargs: P.kwargs):
		builder = ReplyKeyboardBuilder()
		
		rows = function(*args, **kwargs)
		
		for row in rows:
			if len(row) > 10:
				raise ValueError("Reply keyboards can have a maximum of 10 buttons per column")
			
			builder.row(*[KeyboardButton(text=button) for button in row])

		return builder.as_markup()
		
	return wrapper

def inline(function: Callable[P, list[list[tuple[str, str]]]]):
	@wraps(function)
	def wrapper(*args: P.args, **kwargs: P.kwargs):
		builder = InlineKeyboardBuilder()
		
		rows = function(*args, **kwargs)
		
		for row in rows:
			if len(row) > 8:
				raise ValueError("Inline keyboards can have a maximum of 8 buttons per column")
				
			buttons = []

			for button in row:
				if len(button) != 2:
					raise ValueError("InlineKeyboardButton must be a tuple of (text, callback_data)")
				
				text = button[0]
				value = button[1]

				if value.startswith("t.me"):
					value = "https://" + value
				
				if value.startswith(("https://", "http://", "tg://")):
					buttons.append(InlineKeyboardButton(text=text, url=value))
				else:
					buttons.append(InlineKeyboardButton(text=text, callback_data=value))
			
			builder.row(*buttons)

		return builder.as_markup()
		
	return wrapper

__all__ = ["reply", "inline"]


from typing import Any, Optional, Union

from aiogram import Dispatcher, Router
from aiogram.client.default import Default
from aiogram.enums import ChatType, ContentType
from aiogram.types import Message, MessageEntity, LinkPreviewOptions, ReplyParameters, ReplyMarkupUnion

from .middleware import AnalyticsMiddleware
from .models import BotMessageModel
from ..sqlalchemy.session import AsyncSessionManagerFactory
from ..utils.logging import loggers

async def _empty_callback(event):
	pass

async def _false_filter(event):
	return False

async def setup_analytics(
	router: Router,
	path="analytics.db",
	chats: bool=True,
	messages: bool=True,
	bot_messages: bool=True,
	callback_queries: bool=True,
	inline_queries: bool=False
):
	"""
	Setup analytics for `router`.

	:param router: Router to setup analytics for
	:param chats: Whether to track chats
	:param messages: Whether to track messages
	:param bot_messages: Whether to track bot (outgoing) messages
	:param callback_queries: Whether to track callback queries
	:param inline_queries: Whether to track inline queries
	"""

	session_manager_factory = AsyncSessionManagerFactory(path)

	middleware = AnalyticsMiddleware(
		engine=session_manager_factory.engine,
		chats=chats,
		messages=messages,
		callback_queries=callback_queries,
		inline_queries=inline_queries
	)

	await middleware.create_all()

	router.my_chat_member.register(_empty_callback, _false_filter)
	router.my_chat_member.outer_middleware(middleware)

	if messages:
		router.message.register(_empty_callback, _false_filter)
		router.message.outer_middleware(middleware)

	if bot_messages:
		super_answer = Message.answer

		async def answer(
			self,
			text: str,
			parse_mode: Optional[Union[str, Default]] = Default("parse_mode"),
			entities: Optional[list[MessageEntity]] = None,
			link_preview_options: Optional[Union[LinkPreviewOptions, Default]] = Default(
				"link_preview"
			),
			disable_notification: Optional[bool] = None,
			protect_content: Optional[Union[bool, Default]] = Default("protect_content"),
			allow_paid_broadcast: Optional[bool] = None,
			message_effect_id: Optional[str] = None,
			reply_parameters: Optional[ReplyParameters] = None,
			reply_markup: Optional[ReplyMarkupUnion] = None,
			allow_sending_without_reply: Optional[bool] = None,
			disable_web_page_preview: Optional[Union[bool, Default]] = Default(
				"link_preview_is_disabled"
			),
			reply_to_message_id: Optional[int] = None,
			**kwargs: Any,
		):
			message = await super_answer(self, text, parse_mode, entities, link_preview_options, disable_notification, protect_content, allow_paid_broadcast, message_effect_id, reply_parameters, reply_markup, allow_sending_without_reply, disable_web_page_preview, reply_to_message_id, **kwargs)

			if message.content_type == ContentType.PHOTO:
				file_id = message.photo[-1].file_id
			elif message.content_type in [
				ContentType.DOCUMENT,
				ContentType.VIDEO,
				ContentType.AUDIO,
				ContentType.VOICE,
				ContentType.VIDEO_NOTE,
				ContentType.ANIMATION,
				ContentType.STICKER
			]:
				file_id = getattr(message, message.content_type).file_id
			else:
				file_id = None

			if message.reply_markup and message.reply_markup.inline_keyboard:
				inline_keyboard = message.model_dump()["reply_markup"]["inline_keyboard"]
			else:
				inline_keyboard = None

			async with session_manager_factory() as session:
				model = BotMessageModel(
					id=message.message_id,
					chat_id=message.chat.id,
					text=message.text or message.caption,
					md_text=message.md_text,
					html_text=message.html_text,
					entities=message.model_dump()["entities"],
					content_type=message.content_type,
					file_id=file_id,
					media_group_id=message.media_group_id,
					reply_to_message_id=message.reply_to_message.message_id if message.reply_to_message else None,
					inline_keyboard=inline_keyboard,
					date=message.date
				)

				session.add(model)

			if message.chat.type == ChatType.PRIVATE:
				chat_type = "user"
			else:
				chat_type = message.chat.type

			loggers.analytics.info(f"New outgoing message to {chat_type} id={message.chat.id}")

			return message
		
		Message.answer = answer

	if callback_queries:
		router.callback_query.register(_empty_callback, _false_filter)
		router.callback_query.outer_middleware(middleware)

	if inline_queries:
		router.inline_query.register(_empty_callback, _false_filter)
		router.inline_query.outer_middleware(middleware)

	if isinstance(router, Dispatcher):
		loggers.analytics.info("Setup analytics for dispatcher")
	else:
		loggers.analytics.info(f"Setup analytics for router name={router.name!r}")

__all__ = ["setup_analytics"]


from .static import (
	reply as static_reply_markup,
	inline as static_inline_markup
)

from .dynamic import (
	reply as dynamic_reply_markup,
	inline as dynamic_inline_markup
)

__all__ = [
	"static_reply_markup", "static_inline_markup",
	"dynamic_reply_markup", "dynamic_inline_markup"
]


from collections.abc import Callable

from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker, AsyncEngine, AsyncSession

class AsyncSessionManager:
	"""
	Asynchronous context manager for handling commitment and rollback of `AsyncSession` objects.

	The session is commited on successful exit.
	If an exception occurs, the transaction is rolled back.
	Finally, the session is closed.

	Example::
	
		AsyncSession = async_sessionmaker(engine)

		async with AsyncSessionManager(AsyncSession) as session:
			await session.execute(update(...))

			# On success, transaction is commited.
			# On exception, it is rolled back.
			# Session is closed regardless.

	:param session_factory: Callable that returns an `AsyncSession`
	"""
	
	def __init__(self, session_factory: Callable[..., AsyncSession]):
		self.session_factory = session_factory
		self.session: AsyncSession | None = None
	
	async def __aenter__(self):
		self.session = self.session_factory()

		return self.session
	
	async def __aexit__(self, exc_type, exc, tb):
		try:
			if exc_type is None:
				await self.session.commit()
			else:
				await self.session.rollback()
		finally:
			await self.session.close()

			self.session = None

class AsyncSessionManagerFactory:
	def __init__(self, engine: str | AsyncEngine=None):
		if isinstance(engine, str):
			if not engine.startswith("sqlite+aiosqlite:///"):
				engine = "sqlite+aiosqlite:///" + engine

			self.engine = create_async_engine(engine)
		elif isinstance(engine, AsyncEngine):
			self.engine = engine

		self.session_factory = async_sessionmaker(self.engine, expire_on_commit=False)

	def __call__(self):
		return AsyncSessionManager(self.session_factory)

__all__ = ["AsyncSessionManager", "AsyncSessionManagerFactory"]


import functools
from typing import Callable
from typing_extensions import TypeVar, ParamSpec

from aiogram.types import ReplyKeyboardMarkup, InlineKeyboardMarkup
from aiogram.utils.keyboard import ReplyKeyboardBuilder, InlineKeyboardBuilder

T = TypeVar("T")
P = ParamSpec("P")

def reply(function: Callable[P, T]):
	@functools.wraps(function)
	def wrapper(*args: P.args, **kwargs: P.kwargs) -> ReplyKeyboardMarkup:
		builder = ReplyKeyboardBuilder()
		
		function(builder, *args, **kwargs)

		return builder.as_markup()
		
	return wrapper

def inline(function: Callable[P, T]):
	@functools.wraps(function)
	def wrapper(*args: P.args, **kwargs: P.kwargs) -> InlineKeyboardMarkup:
		builder = InlineKeyboardBuilder()
		
		function(builder, *args, **kwargs)

		return builder.as_markup()
	
	return wrapper

__all__ = ["reply", "inline"]


from datetime import datetime
from enum import Enum
from typing import Optional

from sqlalchemy import Enum as SQLEnum, ForeignKey, JSON
from sqlalchemy.orm import mapped_column, relationship, DeclarativeBase, Mapped

from ..sqlalchemy.mixins import ReprMixin

class SQLChatType(Enum):
    SENDER = "sender"
    PRIVATE = "private"
    GROUP = "group"
    SUPERGROUP = "supergroup"
    CHANNEL = "channel"

class BaseModel(ReprMixin, DeclarativeBase):
	pass

class ChatModel(BaseModel):
	__tablename__ = "chats"

	id: Mapped[int] = mapped_column(primary_key=True, autoincrement=False)
	type: Mapped[SQLChatType] = mapped_column(SQLEnum(SQLChatType))
	title: Mapped[Optional[str]]
	full_name: Mapped[str]
	username: Mapped[Optional[str]]
	bio: Mapped[Optional[str]]
	joined_at: Mapped[Optional[datetime]]

	messages: Mapped[list["MessageModel"]] = relationship(back_populates="chat")
	bot_messages: Mapped[list["BotMessageModel"]] = relationship(back_populates="chat")

class UserModel(BaseModel):
	__tablename__ = "users"
	
	id: Mapped[int] = mapped_column(primary_key=True, autoincrement=False)
	username: Mapped[Optional[str]]
	first_name: Mapped[str]
	last_name: Mapped[Optional[str]]
	full_name: Mapped[str]
	bio: Mapped[Optional[str]]
	language_code: Mapped[Optional[str]]
	is_premium: Mapped[bool] = mapped_column(default=False)
	joined_at: Mapped[Optional[datetime]]
	blocked: Mapped[bool] = mapped_column(default=False)

	messages: Mapped[list["MessageModel"]] = relationship(back_populates="user")
	callback_queries: Mapped[list["CallbackQueryModel"]] = relationship(back_populates="user")
	inline_queries: Mapped[list["InlineQueryModel"]] = relationship(back_populates="user")

class MessageModelMixin:
	id: Mapped[int] = mapped_column(primary_key=True, autoincrement=False)
	chat_id: Mapped[int] = mapped_column(ForeignKey("chats.id"))
	text: Mapped[Optional[str]]
	md_text: Mapped[Optional[str]]
	html_text: Mapped[Optional[str]]
	entities: Mapped[Optional[list[dict]]] = mapped_column(JSON)
	content_type: Mapped[str]
	file_id: Mapped[Optional[str]]
	media_group_id: Mapped[Optional[str]]
	reply_to_message_id: Mapped[Optional[int]]
	date: Mapped[datetime]

class MessageModel(MessageModelMixin, BaseModel):
	__tablename__ = "messages"

	chat: Mapped["ChatModel"] = relationship(back_populates="messages")
	user_id: Mapped[int] = mapped_column(ForeignKey("users.id"))
	user: Mapped["UserModel"] = relationship(back_populates="messages")

class BotMessageModel(MessageModelMixin, BaseModel):
	__tablename__ = "bot_messages"
	
	chat: Mapped["ChatModel"] = relationship(back_populates="bot_messages")
	inline_keyboard: Mapped[Optional[list[list[dict]]]] = mapped_column(JSON)

class CallbackQueryModel(BaseModel):
	__tablename__ = "callback_queries"
	
	id: Mapped[str] = mapped_column(primary_key=True)
	user_id: Mapped[int] = mapped_column(ForeignKey("users.id"))
	user: Mapped["UserModel"] = relationship(back_populates="callback_queries")
	message_id: Mapped[int] = mapped_column(ForeignKey("bot_messages.id"))
	message: Mapped["BotMessageModel"] = relationship()
	data: Mapped[str]
	date: Mapped[datetime]

class InlineQueryModel(BaseModel):
	__tablename__ = "inline_queries"
	
	id: Mapped[int] = mapped_column(primary_key=True, autoincrement=False)
	user_id: Mapped[int] = mapped_column(ForeignKey("users.id"))
	user: Mapped["UserModel"] = relationship(back_populates="inline_queries")
	query: Mapped[str]
	chat_type: Mapped[str]
	date: Mapped[datetime]

__all__ = ["BaseModel", "ChatModel","UserModel", "MessageModel", "BotMessageModel", "InlineQueryModel"]


from aiogram import Router
from aiogram.enums import ParseMode
from aiogram.filters import Command
from aiogram.types import Message, CallbackQuery, InlineKeyboardMarkup, InlineKeyboardButton

from tgtilities.sqlalchemy.session import AsyncSessionManagerFactory
from tgtilities.utils.errors import SupressNotModified

router = Router(name="test")

@router.message(Command("add"))
async def add(message: Message, session_factory: AsyncSessionManagerFactory):
	await message.answer()

@router.callback_query()
async def callback_query(callback: CallbackQuery):
	button = [button for buttons in callback.message.reply_markup.inline_keyboard for button in buttons if button.callback_data == callback.data][0]

	with SupressNotModified():
		await callback.message.edit_text(
			f"You just hit the <code>{callback.data}</code> button! Click it again 👇",
			reply_markup=InlineKeyboardMarkup(inline_keyboard=[[InlineKeyboardButton(text=button.text, callback_data=button.callback_data)]]),
			parse_mode=ParseMode.HTML
		)

	await callback.answer()